package main

import "fmt"
import "math"
import "os"
import "time"

var globalErrorCount int = 33
var globalErr = fmt.Errorf("error")
var iter int = 1000000000

type test struct {
	fn func(string)
	arg string
	name string
	duration time.Duration
}

func run(t *test) {
	var i int
	var start time.Time

	start = time.Now()
	for i = 0; i < iter; i++ {
		t.fn(t.arg)
	}
	t.duration = time.Since(start)
}

func digit3(number float64) string {
	if number < 10 {
		return fmt.Sprintf("%.2f", math.Round(number*100)/100)
	} else if number < 100 {
		return fmt.Sprintf("%.1f", math.Round(number*10)/10)
	} else {
		return fmt.Sprintf("%.0f", math.Round(number))
	}
}

func results(ts []*test) {
	var faster time.Duration
	var t *test

	faster = 99 * 24 * 60 * 60 * time.Second
	for _, t = range ts {
		if t.duration < faster {
			faster = t.duration
		}
	}

	for _, t = range ts {
		if t.duration == faster {
			fmt.Printf("%s(%q) %ss\n", t.name, t.arg, digit3(t.duration.Seconds()))
		} else {
			fmt.Printf("%s(%q) %ss %s%%\n",  t.name, t.arg, digit3(t.duration.Seconds()),
				digit3(float64(t.duration - faster) / float64(faster) * 100.0))
		}
	}
}

func main() {
	var ts [3]*test

	ts = [3]*test{
		&test{
			fn: test_closure,
			arg: os.Args[1],
			name: "test_closure",
		},
		&test{
			fn: test_lambda,
			arg: os.Args[1],
			name: "test_lambda",
		},
		&test{
			fn: test_goto,
			arg: os.Args[1],
			name: "test_goto",
		},
	}

	fmt.Printf("run test 1\n")
	run(ts[0])
	fmt.Printf("run test 2\n")
	run(ts[1])
	fmt.Printf("run test 3\n")
	run(ts[2])
	results(ts[:])

	ts = [3]*test{
		&test{
			fn: test_closure,
			arg: os.Args[2],
			name: "test_closure",
		},
		&test{
			fn: test_lambda,
			arg: os.Args[2],
			name: "test_lambda",
		},
		&test{
			fn: test_goto,
			arg: os.Args[2],
			name: "test_goto",
		},
	}

	fmt.Printf("run test 1\n")
	run(ts[0])
	fmt.Printf("run test 2\n")
	run(ts[1])
	fmt.Printf("run test 3\n")
	run(ts[2])
	results(ts[:])

	println(globalErrorCount)
}

func test_closure(str string) {
	var err error

	func() {
		if len(str) > 3 {
			err = globalErr
			return
		}
		if len(str) > 4 {
			err = globalErr
			return
		}
	} ()

	if err != nil {
		globalErrorCount++
	}
}

func test_lambda(str string) {
	var err error

	err = func(str_test string) error {
		if len(str_test) > 3 {
			return globalErr
		}
		if len(str_test) > 4 {
			return globalErr
		}
		return nil
	} (str)

	if err != nil {
		globalErrorCount++
	}
}

func test_goto(str string) {
	var err error

	if len(str) > 3 {
		err = globalErr
		goto process_error
	}
	if len(str) > 4 {
		err = globalErr
		goto process_error
	}

process_error:

	if err != nil {
		globalErrorCount++
	}
}
